"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.traverseMemberExpressionUtil = exports.findParentOfType = exports.findParentVariableDeclaration = exports.findParentCallExpression = exports.getMemberExpressionElements = exports.getIdentifierFromExpression = void 0;
/**
 * Find the identifier from a given MemberExpression
 *
 * Example: return `foo` for a node of `foo.bar.baz()`
 *
 * @param  {MemberExpression} node
 * @return {string|null}
 */
function getIdentifierFromExpression(node) {
    if (!node) {
        return null;
    }
    if (node.type === 'Identifier') {
        return node;
    }
    return getIdentifierFromExpression(node.object);
}
exports.getIdentifierFromExpression = getIdentifierFromExpression;
/**
 * Returns list of elements from a given MemberExpression
 *
 * Example: return ['foo', 'bar', 'baz'] for a node of `foo.bar.baz()`
 *
 * @param  {MemberExpression} node
 * @return {array}
 */
function getMemberExpressionElements(node, _rest) {
    if (_rest === void 0) { _rest = []; }
    if (node.object.type === 'Identifier') {
        return [node.object.name, node.property.name].concat(_rest);
    }
    return getMemberExpressionElements(node.object, [node.property.name].concat(_rest));
}
exports.getMemberExpressionElements = getMemberExpressionElements;
function findParentCallExpression(path, name) {
    if (!path) {
        return null;
    }
    if (path.value.type === 'CallExpression' &&
        path.value.callee.property &&
        path.value.callee.property.name === name) {
        return path;
    }
    return findParentCallExpression(path.parentPath, name);
}
exports.findParentCallExpression = findParentCallExpression;
function findParentVariableDeclaration(path) {
    return findParentOfType(path, 'VariableDeclarator');
}
exports.findParentVariableDeclaration = findParentVariableDeclaration;
function findParentOfType(path, type) {
    if (!path || !path.value) {
        return null;
    }
    if (path.value.type === type) {
        return path;
    }
    return findParentOfType(path.parentPath, type);
}
exports.findParentOfType = findParentOfType;
function traverseMemberExpressionUtil(j, nodeValidator) {
    var traverseMemberExpression = function (node) {
        if (!node) {
            return false;
        }
        if (nodeValidator(node)) {
            return true;
        }
        return traverseMemberExpression(node.object);
    };
    return traverseMemberExpression;
}
exports.traverseMemberExpressionUtil = traverseMemberExpressionUtil;
